import { Subject, ReplaySubject, isObservable, NEVER, from } from 'rxjs';
import { mergeAll, share, materialize, tap, map, distinctUntilChanged, startWith, filter, switchMap } from 'rxjs/operators';
import { coerceAllFactory } from '@rx-angular/cdk/coercing';

function toRxErrorNotification(error, value) {
    return {
        value,
        kind: "error" /* Error */,
        hasValue: !!value || false,
        complete: false,
        error: error || true,
    };
}
function toRxSuspenseNotification(value) {
    return {
        value,
        kind: "suspense" /* Suspense */,
        hasValue: !!value || false,
        complete: false,
        error: false,
    };
}
function toRxCompleteNotification(value) {
    return {
        value,
        kind: "complete" /* Complete */,
        hasValue: !!value || false,
        complete: true,
        error: false,
    };
}

/**
 * @internal
 *
 * A factory function returning an object to handle the process of switching templates by Notification channel.
 * You can next a Observable of `RxNotification` multiple times and merge them into the Observable exposed under `trigger$`
 *
 */
function templateTriggerHandling() {
    const hotFlattened = coerceAllFactory(() => new Subject(), mergeAll());
    return {
        next(templateName) {
            hotFlattened.next(templateName);
        },
        trigger$: hotFlattened.values$.pipe(share()),
    };
}

function rxMaterialize() {
    return (o$) => o$.pipe(materialize(), tap(({ kind, error }) => {
        if (kind === 'E') {
            console.error(error);
        }
    }), map(({ value, hasValue, error, kind }) => {
        const rxNotificationKind = notificationKindToRxNotificationKind(kind);
        return {
            value,
            hasValue,
            error,
            kind: rxNotificationKind,
            complete: rxNotificationKind === "complete" /* Complete */,
        };
    }));
}
function notificationKindToRxNotificationKind(kind) {
    switch (kind) {
        case 'C':
            return "complete" /* Complete */;
        case 'E':
            return "error" /* Error */;
        case 'N':
        default:
            return "next" /* Next */;
    }
}

/**
 * @internal
 */
function createTemplateNotifier(withSuspense) {
    const observablesSubject = new ReplaySubject(1);
    let firstRun = true;
    let latestNextValue;
    const wSuspense = () => withSuspense && withSuspense();
    const values$ = observablesSubject.pipe(distinctUntilChanged(), 
    // handle null | undefined assignment and new Observable reset
    map((observable$) => {
        const isObs = isObservable(observable$);
        // only pass through initial undefined value to filter out
        if (firstRun && (observable$ === undefined || observable$ === NEVER)) {
            return wSuspense() ? NEVER.pipe(startWith(undefined)) : undefined;
        }
        const isNull = observable$ == null;
        if (isNull) {
            return NEVER.pipe(startWith(observable$));
        }
        return (isObs ? observable$ : from(observable$)).pipe((o$) => {
            if (!firstRun || (withSuspense && withSuspense())) {
                return o$.pipe(startWith(undefined));
            }
            return o$;
        });
    }), tap(() => (firstRun = false)), filter((v) => v !== undefined), switchMap((o) => {
        return o.pipe(distinctUntilChanged(), rxMaterialize(), map((notification) => {
            latestNextValue =
                notification.kind === "next" /* Next */
                    ? notification.value
                    : latestNextValue;
            if (notification.kind === "next" /* Next */ &&
                latestNextValue === undefined) {
                return toRxSuspenseNotification(latestNextValue);
            }
            return Object.assign(Object.assign({}, notification), { value: latestNextValue });
        }));
    }));
    return {
        next(observable) {
            observablesSubject.next(observable);
        },
        values$,
    };
}

/**
 * Generated bundle index. Do not edit.
 */

export { createTemplateNotifier, rxMaterialize, templateTriggerHandling, toRxCompleteNotification, toRxErrorNotification, toRxSuspenseNotification };
//# sourceMappingURL=cdk-notifications.js.map
